@@ -3,49 +3,64 @@
 ## Copyright (C) Philippe Biondi <phil@secdev.org>
 ## This program is published under a GPLv2 license
 
-###############
+"""
+Generators and packet meta classes.
+"""
+
+################
 ## Generators ##
 ################
 
+from __future__ import absolute_import
 import re,random,socket
-import config
-import error
+import types
+from scapy.modules.six.moves import range
 
 class Gen(object):
+    __slots__ = []
     def __iter__(self):
         return iter([])
-    
+
+
+def _get_values(value):
+    """Generate a range object from (start, stop[, step]) tuples, or
+    return value.
+
+    """
+    if (isinstance(value, tuple) and (2 <= len(value) <= 3) and \
+        all(hasattr(i, "__int__") for i in value)):
+        # We use values[1] + 1 as stop value for (x)range to maintain
+        # the behavior of using tuples as field `values`
+        return range(*((int(value[0]), int(value[1]) + 1)
+                       + tuple(int(v) for v in value[2:])))
+    return value
+
+
 class SetGen(Gen):
-    def __init__(self, set, _iterpacket=1):
+    def __init__(self, values, _iterpacket=1):
         self._iterpacket=_iterpacket
-        if type(set) is list:
-            self.set = set
-        elif isinstance(set, BasePacketList):
-            self.set = list(set)
+        if isinstance(values, (list, BasePacketList)):
+            self.values = [_get_values(val) for val in values]
         else:
-            self.set = [set]
+            self.values = [_get_values(values)]
     def transf(self, element):
         return element
     def __iter__(self):
-        for i in self.set:
-            if (type(i) is tuple) and (len(i) == 2) and type(i[0]) is int and type(i[1]) is int:
-                if  (i[0] <= i[1]):
-                    j=i[0]
-                    while j <= i[1]:
-                        yield j
-                        j += 1
-            elif isinstance(i, Gen) and (self._iterpacket or not isinstance(i,BasePacket)):
+        for i in self.values:
+            if (isinstance(i, Gen) and
+                (self._iterpacket or not isinstance(i,BasePacket))) or (
+                    isinstance(i, (range, types.GeneratorType))):
                 for j in i:
                     yield j
             else:
                 yield i
     def __repr__(self):
-        return "<SetGen %s>" % self.set.__repr__()
+        return "<SetGen %r>" % self.values
 
 class Net(Gen):
     """Generate a list of IPs from a network address or a name"""
     name = "ip"
-    ipaddress = re.compile(r"^(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)(/[0-3]?[0-9])?$")
+    ip_regex = re.compile(r"^(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)(/[0-3]?[0-9])?$")
 
     @staticmethod
     def _parse_digit(a,netmask):
@@ -53,39 +68,38 @@ class Net(Gen):
         if a == "*":
             a = (0,256)
         elif a.find("-") >= 0:
-            x,y = map(int,a.split("-"))
+            x, y = [int(d) for d in a.split('-')]
             if x > y:
                 y = x
-            a = (x &  (0xffL<<netmask) , max(y, (x | (0xffL>>(8-netmask))))+1)
+            a = (x &  (0xff<<netmask) , max(y, (x | (0xff>>(8-netmask))))+1)
         else:
-            a = (int(a) & (0xffL<<netmask),(int(a) | (0xffL>>(8-netmask)))+1)
+            a = (int(a) & (0xff<<netmask),(int(a) | (0xff>>(8-netmask)))+1)
         return a
 
     @classmethod
     def _parse_net(cls, net):
         tmp=net.split('/')+["32"]
-        if not cls.ipaddress.match(net):
+        if not cls.ip_regex.match(net):
             tmp[0]=socket.gethostbyname(tmp[0])
         netmask = int(tmp[1])
-        return map(lambda x,y: cls._parse_digit(x,y), tmp[0].split("."), map(lambda x,nm=netmask: x-nm, (8,16,24,32))),netmask
+        ret_list = [cls._parse_digit(x, y-netmask) for (x, y) in zip(tmp[0].split('.'), [8, 16, 24, 32])]
+        return ret_list, netmask
 
     def __init__(self, net):
         self.repr=net
         self.parsed,self.netmask = self._parse_net(net)
 
-
+    def __str__(self):
+        return next(self.__iter__(), None)
                                                                                                
     def __iter__(self):
-        for d in xrange(*self.parsed[3]):
-            for c in xrange(*self.parsed[2]):
-                for b in xrange(*self.parsed[1]):
-                    for a in xrange(*self.parsed[0]):
+        for d in range(*self.parsed[3]):
+            for c in range(*self.parsed[2]):
+                for b in range(*self.parsed[1]):
+                    for a in range(*self.parsed[0]):
                         yield "%i.%i.%i.%i" % (a,b,c,d)
     def choice(self):
-        ip = []
-        for v in self.parsed:
-            ip.append(str(random.randint(v[0],v[1]-1)))
-        return ".".join(ip) 
+        return ".".join(str(random.randint(v[0], v[1] - 1)) for v in self.parsed)
                           
     def __repr__(self):
         return "Net(%r)" % self.repr
@@ -100,10 +114,7 @@ class Net(Gen):
             p2 = other.parsed
         else:
             p2,nm2 = self._parse_net(other)
-        for (a1,b1),(a2,b2) in zip(self.parsed,p2):
-            if a1 > a2 or b1 < b2:
-                return False
-        return True
+        return all(a1 <= a2 and b1 >= b2 for (a1, b1), (a2, b2) in zip(self.parsed, p2))
     def __rcontains__(self, other):        
         return self in self.__class__(other)
         
@@ -125,10 +136,10 @@ class OID(Gen):
         return "OID(%r)" % self.oid
     def __iter__(self):        
         ii = [k[0] for k in self.cmpt]
-        while 1:
+        while True:
             yield self.fmt % tuple(ii)
             i = 0
-            while 1:
+            while True:
                 if i >= len(ii):
                     raise StopIteration
                 if ii[i] < self.cmpt[i][1]:
@@ -155,7 +166,7 @@ class Packet_metaclass(type):
                         resolved_fld.append(f2)
                 else:
                     resolved_fld.append(f)
-        else: # look for a field_desc in parent classes
+        else: # look for a fields_desc in parent classes
             resolved_fld = None
             for b in bases:
                 if hasattr(b,"fields_desc"):
@@ -173,11 +184,27 @@ class Packet_metaclass(type):
 
             dct["fields_desc"] = final_fld
 
+        dct.setdefault("__slots__", [])
+        for attr in ["name", "overload_fields"]:
+            try:
+                dct["_%s" % attr] = dct.pop(attr)
+            except KeyError:
+                pass
         newcls = super(Packet_metaclass, cls).__new__(cls, name, bases, dct)
+        newcls.__all_slots__ = set(
+            attr
+            for cls in newcls.__mro__ if hasattr(cls, "__slots__")
+            for attr in cls.__slots__
+        )
+
+        newcls.aliastypes = [newcls] + getattr(newcls, "aliastypes", [])
+
         if hasattr(newcls,"register_variant"):
             newcls.register_variant()
-        for f in newcls.fields_desc:                
-            f.register_owner(newcls)
+        for f in newcls.fields_desc:
+            if hasattr(f, "register_owner"):
+                f.register_owner(newcls)
+        from scapy import config
         config.conf.layers.register(newcls)
         return newcls
 
@@ -189,11 +216,22 @@ class Packet_metaclass(type):
 
     def __call__(cls, *args, **kargs):
         if "dispatch_hook" in cls.__dict__:
-            cls =  cls.dispatch_hook(*args, **kargs)
+            try:
+                cls = cls.dispatch_hook(*args, **kargs)
+            except:
+                from scapy import config
+                if config.conf.debug_dissector:
+                    raise
+                cls = config.conf.raw_layer
         i = cls.__new__(cls, cls.__name__, cls.__bases__, cls.__dict__)
         i.__init__(*args, **kargs)
         return i
 
+class Field_metaclass(type):
+    def __new__(cls, name, bases, dct):
+        dct.setdefault("__slots__", [])
+        newcls = super(Field_metaclass, cls).__new__(cls, name, bases, dct)
+        return newcls
 
 class NewDefaultValues(Packet_metaclass):
     """NewDefaultValues is deprecated (not needed anymore)
@@ -203,7 +241,7 @@ class NewDefaultValues(Packet_metaclass):
     and it should still work.
     """    
     def __new__(cls, name, bases, dct):
-        from error import log_loading
+        from scapy.error import log_loading
         import traceback
         try:
             for tb in traceback.extract_stack()+[("??",-1,None,"")]:
@@ -213,20 +251,17 @@ class NewDefaultValues(Packet_metaclass):
         except:
             f,l="??",-1
             raise
-        log_loading.warning("Deprecated (no more needed) use of NewDefaultValues  (%s l. %i)." % (f,l))
+        log_loading.warning("Deprecated (no more needed) use of NewDefaultValues  (%s l. %i).", f, l)
         
         return super(NewDefaultValues, cls).__new__(cls, name, bases, dct)
 
 class BasePacket(Gen):
-    pass
+    __slots__ = []
 
 
 #############################
-## Packet list base classe ##
+## Packet list base class  ##
 #############################
 
-class BasePacketList:
-    pass
-
-
-
+class BasePacketList(object):
+    __slots__ = []
