@@ -3,9 +3,16 @@
 ## Copyright (C) Philippe Biondi <phil@secdev.org>
 ## This program is published under a GPLv2 license
 
+"""
+Fields that hold random numbers.
+"""
+
+from __future__ import absolute_import
 import random,time,math
-from base_classes import Net
-from utils import corrupt_bits,corrupt_bytes
+from scapy.base_classes import Net
+from scapy.compat import *
+from scapy.utils import corrupt_bits,corrupt_bytes
+from scapy.modules.six.moves import range
 
 ####################
 ## Random numbers ##
@@ -31,7 +38,7 @@ class RandomEnumeration:
             n += 1
         self.n =n
 
-        self.fs = min(3,(n+1)/2)
+        self.fs = min(3,(n+1)//2)
         self.fsmask = 2**self.fs-1
         self.rounds = max(self.n,3)
         self.turns = 0
@@ -43,12 +50,13 @@ class RandomEnumeration:
         while True:
             if self.turns == 0 or (self.i == 0 and self.renewkeys):
                 self.cnt_key = self.rnd.randint(0,2**self.n-1)
-                self.sbox = [self.rnd.randint(0,self.fsmask) for k in xrange(self.sbox_size)]
+                self.sbox = [self.rnd.randint(0, self.fsmask)
+                             for _ in range(self.sbox_size)]
             self.turns += 1
             while self.i < 2**self.n:
                 ct = self.i^self.cnt_key
                 self.i += 1
-                for k in range(self.rounds): # Unbalanced Feistel Network
+                for _ in range(self.rounds): # Unbalanced Feistel Network
                     lsb = ct & self.fsmask
                     ct >>= self.fs
                     lsb ^= self.sbox[ct%self.sbox_size]
@@ -59,22 +67,29 @@ class RandomEnumeration:
             self.i = 0
             if not self.forever:
                 raise StopIteration
+    __next__ = next
 
 
-class VolatileValue:
+class VolatileValue(object):
     def __repr__(self):
         return "<%s>" % self.__class__.__name__
+    def __eq__(self, other):
+        x = self._fix()
+        y = other._fix() if isinstance(other, VolatileValue) else other
+        if not isinstance(x, type(y)):
+            return False
+        return x == y
     def __getattr__(self, attr):
-        if attr == "__setstate__":
+        if attr in ["__setstate__", "__getstate__"]:
             raise AttributeError(attr)
-        elif attr == "__cmp__":
-            x = self._fix()
-            def cmp2(y,x=x):
-                if type(x) != type(y):
-                    return -1
-                return x.__cmp__(y)
-            return cmp2
         return getattr(self._fix(),attr)
+    def __str__(self):
+        return str(self._fix())
+    def __bytes__(self):
+        return raw(self._fix())
+    def __len__(self):
+        return len(self._fix())
+
     def _fix(self):
         return None
 
@@ -92,21 +107,70 @@ class RandNum(RandField):
     def _fix(self):
         return random.randrange(self.min, self.max+1)
 
-class RandNumGamma(RandField):
+    def __int__(self):
+        return int(self._fix())
+    def __index__(self):
+        return int(self)
+    def __nonzero__(self):
+        return bool(self.value)
+    __bool__ = __nonzero__
+    def __add__(self, other):
+        return self._fix() + other
+    def __radd__(self, other):
+        return other + self._fix()
+    def __sub__(self, other):
+        return self._fix() - other
+    def __rsub__(self, other):
+        return other - self._fix()
+    def __mul__(self, other):
+        return self._fix() * other
+    def __rmul__(self, other):
+        return other * self._fix()
+    def __floordiv__(self, other):
+        return self._fix() / other
+    __div__ = __floordiv__
+
+    def __lt__(self, other):
+        return self._fix() < other
+    def __le__(self, other):
+        return self._fix() <= other
+    def __eq__(self, other):
+        return self._fix() == other
+    def __ne__(self, other):
+        return self._fix() != other
+    def __ge__(self, other):
+        return self._fix() >= other
+    def __gt__(self, other):
+        return self._fix() > other
+
+    def __lshift__(self, other):
+        return self._fix() << other
+    def __rshift__(self, other):
+        return self._fix() >> other
+    def __and__(self, other):
+        return self._fix() & other
+    def __rand__(self, other):
+        return other & self._fix()
+    def __or__(self, other):
+        return self._fix() | other
+    def __ror__(self, other):
+        return other | self._fix()
+
+class RandNumGamma(RandNum):
     def __init__(self, alpha, beta):
         self.alpha = alpha
         self.beta = beta
     def _fix(self):
         return int(round(random.gammavariate(self.alpha, self.beta)))
 
-class RandNumGauss(RandField):
+class RandNumGauss(RandNum):
     def __init__(self, mu, sigma):
         self.mu = mu
         self.sigma = sigma
     def _fix(self):
         return int(round(random.gauss(self.mu, self.sigma)))
 
-class RandNumExpo(RandField):
+class RandNumExpo(RandNum):
     def __init__(self, lambd, base=0):
         self.lambd = lambd
         self.base = base
@@ -115,74 +179,83 @@ class RandNumExpo(RandField):
 
 class RandEnum(RandNum):
     """Instances evaluate to integer sampling without replacement from the given interval"""
-    def __init__(self, min, max):
-        self.seq = RandomEnumeration(min,max)
+    def __init__(self, min, max, seed=None):
+        self.seq = RandomEnumeration(min,max,seed)
     def _fix(self):
-        return self.seq.next()
+        return next(self.seq)
 
 class RandByte(RandNum):
     def __init__(self):
-        RandNum.__init__(self, 0, 2L**8-1)
+        RandNum.__init__(self, 0, 2**8-1)
 
 class RandSByte(RandNum):
     def __init__(self):
-        RandNum.__init__(self, -2L**7, 2L**7-1)
+        RandNum.__init__(self, -2**7, 2**7-1)
 
 class RandShort(RandNum):
     def __init__(self):
-        RandNum.__init__(self, 0, 2L**16-1)
+        RandNum.__init__(self, 0, 2**16-1)
 
 class RandSShort(RandNum):
     def __init__(self):
-        RandNum.__init__(self, -2L**15, 2L**15-1)
+        RandNum.__init__(self, -2**15, 2**15-1)
 
 class RandInt(RandNum):
     def __init__(self):
-        RandNum.__init__(self, 0, 2L**32-1)
+        RandNum.__init__(self, 0, 2**32-1)
 
 class RandSInt(RandNum):
     def __init__(self):
-        RandNum.__init__(self, -2L**31, 2L**31-1)
+        RandNum.__init__(self, -2**31, 2**31-1)
 
 class RandLong(RandNum):
     def __init__(self):
-        RandNum.__init__(self, 0, 2L**64-1)
+        RandNum.__init__(self, 0, 2**64-1)
 
 class RandSLong(RandNum):
     def __init__(self):
-        RandNum.__init__(self, -2L**63, 2L**63-1)
+        RandNum.__init__(self, -2**63, 2**63-1)
 
 class RandEnumByte(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, 0, 2L**8-1)
+        RandEnum.__init__(self, 0, 2**8-1)
 
 class RandEnumSByte(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, -2L**7, 2L**7-1)
+        RandEnum.__init__(self, -2**7, 2**7-1)
 
 class RandEnumShort(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, 0, 2L**16-1)
+        RandEnum.__init__(self, 0, 2**16-1)
 
 class RandEnumSShort(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, -2L**15, 2L**15-1)
+        RandEnum.__init__(self, -2**15, 2**15-1)
 
 class RandEnumInt(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, 0, 2L**32-1)
+        RandEnum.__init__(self, 0, 2**32-1)
 
 class RandEnumSInt(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, -2L**31, 2L**31-1)
+        RandEnum.__init__(self, -2**31, 2**31-1)
 
 class RandEnumLong(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, 0, 2L**64-1)
+        RandEnum.__init__(self, 0, 2**64-1)
 
 class RandEnumSLong(RandEnum):
     def __init__(self):
-        RandEnum.__init__(self, -2L**63, 2L**63-1)
+        RandEnum.__init__(self, -2**63, 2**63-1)
+
+class RandEnumKeys(RandEnum):
+    """Picks a random value from dict keys list. """
+    def __init__(self, enum, seed=None):
+        self.enum = list(enum)
+        self.seq = RandomEnumeration(0, len(self.enum) - 1, seed)
+
+    def _fix(self):
+        return self.enum[next(self.seq)]
 
 class RandChoice(RandField):
     def __init__(self, *args):
@@ -193,29 +266,33 @@ class RandChoice(RandField):
         return random.choice(self._choice)
     
 class RandString(RandField):
-    def __init__(self, size=None, chars="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"):
+    def __init__(self, size=None, chars=b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"):
         if size is None:
             size = RandNumExpo(0.01)
         self.size = size
         self.chars = chars
     def _fix(self):
-        s = ""
-        for i in range(self.size):
-            s += random.choice(self.chars)
+        s = b""
+        for _ in range(self.size):
+            s += chb(random.choice(self.chars))
         return s
+    def __str__(self):
+        return plain_str(self._fix())
+    def __bytes__(self):
+        return raw(self._fix())
+    def __mul__(self, n):
+        return self._fix()*n
 
 class RandBin(RandString):
     def __init__(self, size=None):
-        RandString.__init__(self, size, "".join(map(chr,range(256))))
+        super(RandBin, self).__init__(size=size, chars=b"".join(chb(c) for c in range(256)))
 
-
-class RandTermString(RandString):
+class RandTermString(RandBin):
     def __init__(self, size, term):
-        RandString.__init__(self, size, "".join(map(chr,range(1,256))))
-        self.term = term
+        self.term = raw(term)
+        super(RandTermString, self).__init__(size=size)
     def _fix(self):
-        return RandString._fix(self)+self.term
-    
+        return RandBin._fix(self)+self.term
     
 
 class RandIP(RandString):
@@ -279,12 +356,18 @@ class RandIP6(RandString):
                     remain = random.randint(0,remain)
                 for j in range(remain):
                     ip.append("%04x" % random.randint(0,65535))
+            elif isinstance(n, RandNum):
+                ip.append("%04x" % n)
+            elif n == 0:
+              ip.append("0")
             elif not n:
                 ip.append("")
             else:
                 ip.append("%04x" % n)
         if len(ip) == 9:
             ip.remove("")
+        if ip[-1] == "":
+          ip[-1] = "0"
         return ":".join(ip)
 
 class RandOID(RandString):
@@ -305,15 +388,15 @@ class RandOID(RandString):
             return "<%s [%s]>" % (self.__class__.__name__, self.ori_fmt)
     def _fix(self):
         if self.fmt is None:
-            return ".".join(map(str, [self.idnum for i in xrange(1+self.depth)]))
+            return ".".join(str(self.idnum) for _ in range(1 + self.depth))
         else:
             oid = []
             for i in self.fmt:
                 if i == "*":
                     oid.append(str(self.idnum))
                 elif i == "**":
-                    oid += map(str, [self.idnum for i in xrange(1+self.depth)])
-                elif type(i) is tuple:
+                    oid += [str(self.idnum) for i in range(1 + self.depth)]
+                elif isinstance(i, tuple):
                     oid.append(str(random.randrange(*i)))
                 else:
                     oid.append(i)
@@ -342,11 +425,11 @@ class RandRegExp(RandField):
             else:
                 c1 = s[p-1]
                 c2 = s[p+1]
-                rng = "".join(map(chr, range(ord(c1),ord(c2)+1)))
+                rng = "".join(map(chr, range(ord(c1), ord(c2)+1)))
                 s = s[:p-1]+rng+s[p+1:]
         res = m+s
         if invert:
-            res = "".join([chr(x) for x in xrange(256) if chr(x) not in res])
+            res = "".join(chr(x) for x in range(256) if chr(x) not in res)
         return res
 
     @staticmethod
@@ -354,7 +437,7 @@ class RandRegExp(RandField):
         r = ""
         mul = 1
         for e in lst:
-            if type(e) is list:
+            if isinstance(e, list):
                 if mul != 1:
                     mul = mul-1
                     r += RandRegExp.stack_fix(e[1:]*mul, index)
@@ -365,7 +448,7 @@ class RandRegExp(RandField):
                         index[i] = f
                 r += f
                 mul = 1
-            elif type(e) is tuple:
+            elif isinstance(e, tuple):
                 kind,val = e
                 if kind == "cite":
                     r += index[val-1]
@@ -404,7 +487,7 @@ class RandRegExp(RandField):
             elif c == '|':
                 p = current[0]
                 ch = p[-1]
-                if type(ch) is not tuple:
+                if not isinstance(ch, tuple):
                     ch = ("choice",[current])
                     p[-1] = ch
                 else:
@@ -412,7 +495,7 @@ class RandRegExp(RandField):
                 current = [p]
             elif c == ')':
                 ch = current[0][-1]
-                if type(ch) is tuple:
+                if isinstance(ch, tuple):
                     ch[1].append(current)
                 index.append(current)
                 current = current[0]
@@ -463,7 +546,7 @@ class RandRegExp(RandField):
                 if random.randint(0,1):
                     current.pop()
             elif c == '.':
-                current.append(RandChoice(*[chr(x) for x in xrange(256)]))
+                current.append(RandChoice(*[chr(x) for x in range(256)]))
             elif c == '$' or c == '^':
                 pass
             else:
@@ -486,10 +569,10 @@ class RandSingNum(RandSingularity):
             end = -end
             sign = -1
         end_n = int(math.log(end)/math.log(2))+1
-        return set([sign*2**i for i in range(end_n)])            
+        return {sign*2**i for i in range(end_n)}
         
     def __init__(self, mn, mx):
-        sing = set([0, mn, mx, int((mn+mx)/2)])
+        sing = {0, mn, mx, int((mn+mx)/2)}
         sing |= self.make_power_of_two(mn)
         sing |= self.make_power_of_two(mx)
         for i in sing.copy():
@@ -499,39 +582,40 @@ class RandSingNum(RandSingularity):
             if not mn <= i <= mx:
                 sing.remove(i)
         self._choice = list(sing)
+        self._choice.sort()
         
 
 class RandSingByte(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, 0, 2L**8-1)
+        RandSingNum.__init__(self, 0, 2**8-1)
 
 class RandSingSByte(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, -2L**7, 2L**7-1)
+        RandSingNum.__init__(self, -2**7, 2**7-1)
 
 class RandSingShort(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, 0, 2L**16-1)
+        RandSingNum.__init__(self, 0, 2**16-1)
 
 class RandSingSShort(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, -2L**15, 2L**15-1)
+        RandSingNum.__init__(self, -2**15, 2**15-1)
 
 class RandSingInt(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, 0, 2L**32-1)
+        RandSingNum.__init__(self, 0, 2**32-1)
 
 class RandSingSInt(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, -2L**31, 2L**31-1)
+        RandSingNum.__init__(self, -2**31, 2**31-1)
 
 class RandSingLong(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, 0, 2L**64-1)
+        RandSingNum.__init__(self, 0, 2**64-1)
 
 class RandSingSLong(RandSingNum):
     def __init__(self):
-        RandSingNum.__init__(self, -2L**63, 2L**63-1)
+        RandSingNum.__init__(self, -2**63, 2**63-1)
 
 class RandSingString(RandSingularity):
     def __init__(self):
@@ -546,10 +630,10 @@ class RandSingString(RandSingularity):
                          "%",
                          "%%%",
                          "A"*4096,
-                         "\x00"*4096,
-                         "\xff"*4096,
-                         "\x7f"*4096,
-                         "\x80"*4096,
+                         b"\x00"*4096,
+                         b"\xff"*4096,
+                         b"\x7f"*4096,
+                         b"\x80"*4096,
                          " "*4096,
                          "\\"*4096,
                          "("*4096,
@@ -566,7 +650,7 @@ class RandSingString(RandSingularity):
                          "$(reboot)",
                          "`reboot`",
                          "index.php%00",
-                         "\x00",
+                         b"\x00",
                          "%00",
                          "\\",
                          "../../../../../../../../../../../../../../../../../etc/passwd",
@@ -589,6 +673,11 @@ class RandSingString(RandSingularity):
                          r"\\myserver\share",
                          "foo.exe:",
                          "foo.exe\\", ]
+
+    def __str__(self):
+        return str(self._fix())
+    def __bytes__(self):
+        return raw(self._fix())
                              
 
 class RandPool(RandField):
@@ -597,7 +686,7 @@ class RandPool(RandField):
         pool = []
         for p in args:
             w = 1
-            if type(p) is tuple:
+            if isinstance(p, tuple):
                 p,w = p
             pool += [p]*w
         self._pool = pool
@@ -622,14 +711,23 @@ class IntAutoTime(AutoTime):
 
 
 class ZuluTime(AutoTime):
-    def __init__(self, diff=None):
-        self.diff=diff
+    def __init__(self, diff=0):
+        self.diff = diff
     def _fix(self):
-        return time.strftime("%y%m%d%H%M%SZ",time.gmtime(time.time()+self.diff))
+        return time.strftime("%y%m%d%H%M%SZ",
+                             time.gmtime(time.time() + self.diff))
+
+
+class GeneralizedTime(AutoTime):
+    def __init__(self, diff=0):
+        self.diff = diff
+    def _fix(self):
+        return time.strftime("%Y%m%d%H%M%SZ",
+                             time.gmtime(time.time() + self.diff))
 
 
 class DelayedEval(VolatileValue):
-    """ Exemple of usage: DelayedEval("time.time()") """
+    """ Example of usage: DelayedEval("time.time()") """
     def __init__(self, expr):
         self.expr = expr
     def _fix(self):
